@@ -1,6 +1,6 @@
 /* -*- Mode: C; tab-width: 8; indent-tabs-mode: nil; c-basic-offset: 8 -*-
  *
- * Copyright (C) 2004-2005 William Jon McCann <mccann@jhu.edu>
+ * Copyright (C) 2004-2006 William Jon McCann <mccann@jhu.edu>
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -35,8 +35,7 @@ G_BEGIN_DECLS
 typedef enum {
         GS_MODE_BLANK_ONLY,
         GS_MODE_RANDOM,
-        GS_MODE_SINGLE,
-        GS_MODE_DONT_BLANK
+        GS_MODE_SINGLE
 } GSSaverMode;
 
 typedef struct GSPrefsPrivate GSPrefsPrivate;
@@ -50,37 +49,20 @@ typedef struct
         gboolean         debug;
         gboolean         verbose;		/* whether to print out lots of status info */
 
-        gboolean         lock;			/* whether to lock as well as save */
-
-        gboolean         fade;			/* whether to fade to black, if possible */
-        gboolean         unfade;		/* whether to fade from black, if possible */
-
-        GSList          *themes;       		/* the screensaver themes to run */
-
-        GSSaverMode      mode;			/* hack-selection mode */
+        gboolean         idle_activation_enabled; /* whether to activate when idle */
+        gboolean         lock_enabled;		/* whether to lock when active */
+        gboolean         logout_enabled;	/* Whether to offer the logout option */
+        gboolean         user_switch_enabled;	/* Whether to offer the user switch option */
 
         guint            timeout;	       	/* how much idle time before activation */
         guint            lock_timeout;		/* how long after activation locking starts */
-        guint            cycle;			/* how long each hack should run */
-
-        guint            passwd_timeout;	/* how much time before pw dialog goes down */
-        guint            pointer_timeout;	/* how often to check mouse position */
-        guint            notice_events_timeout;	/* how long after window creation to select */
-        guint            watchdog_timeout;     	/* how often to re-raise and re-blank screen */
-
-        gboolean         dpms_enabled;		/* Whether to power down the monitor */
-        guint            dpms_standby;		/* how long until monitor goes black */
-        guint            dpms_suspend;		/* how long until monitor power-saves */
-        guint            dpms_off;		/* how long until monitor powers down */
-
-        gboolean         logout_enabled;	/* Whether to offer the logout option */
-        gboolean         user_switch_enabled;	/* Whether to offer the user switch option */
         guint            logout_timeout;	/* how long until the logout option appears */
+        guint            cycle;			/* how long each theme should run */
 
-        gboolean         use_xidle_extension;	/* which extension to use, if possible */
-        gboolean         use_mit_saver_extension;
-        gboolean         use_sgi_saver_extension;
-        gboolean         use_proc_interrupts;
+        char            *logout_command;        /* command to use to logout */
+
+        GSList          *themes;       		/* the screensaver themes to run */
+        GSSaverMode      mode;			/* theme selection mode */
 } GSPrefs;
 
 typedef struct
