@@ -3,7 +3,7 @@
  * Portions derived from xscreensaver,
  * Copyright (c) 1991-2004 Jamie Zawinski <jwz@jwz.org>
  *
- * Copyright (C) 2004-2005 William Jon McCann <mccann@jhu.edu>
+ * Copyright (C) 2004-2006 William Jon McCann <mccann@jhu.edu>
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -43,8 +43,9 @@
 #include <gdk/gdkx.h>
 
 #include "gnome-screensaver.h"
-#include "gs-watcher-x11.h"
+#include "gs-watcher.h"
 #include "gs-marshal.h"
+#include "gs-debug.h"
 
 static void     gs_watcher_class_init (GSWatcherClass *klass);
 static void     gs_watcher_init       (GSWatcher      *watcher);
@@ -54,9 +55,9 @@ static void     initialize_server_extensions (GSWatcher *watcher);
 
 static gboolean check_pointer_timer          (GSWatcher      *watcher);
 static void     schedule_wakeup_event        (GSWatcher      *watcher,
-                                              int             when,
-                                              gboolean        verbose);
+                                              int             when);
 static gboolean watchdog_timer               (GSWatcher      *watcher);
+static gboolean idle_timer                   (GSWatcher      *watcher);
 
 #define GS_WATCHER_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE ((o), GS_TYPE_WATCHER, GSWatcherPrivate))
 
@@ -72,8 +73,7 @@ struct _PointerPosition
 struct GSWatcherPrivate
 {
         /* settings */
-        guint           verbose : 1;
-        guint           debug : 1;
+        guint           enabled : 1;
         guint           timeout;
         guint           pointer_timeout;
 
@@ -81,13 +81,17 @@ struct GSWatcherPrivate
         guint           use_sgi_saver_extension : 1;
         guint           use_mit_saver_extension : 1;
 
+        guint           notice_timeout;
+
         /* state */
         guint            active : 1;
+        guint            idle : 1;
+        guint            idle_notice : 1;
 
         PointerPosition *poll_position;
 
-        time_t          last_wall_clock_time;
-        time_t          last_activity_time;
+        GTimer         *idle_timer;
+        GTimer         *jump_timer;
         guint           emergency_lock : 1;
 
         guint           timer_id;
@@ -111,13 +115,15 @@ struct GSWatcherPrivate
 };
 
 enum {
-        IDLE,
+        IDLE_CHANGED,
+        IDLE_NOTICE_CHANGED,
         LAST_SIGNAL
 };
 
 enum {
         PROP_0,
-        PROP_TIMEOUT
+        PROP_TIMEOUT,
+        PROP_DEBUG
 };
 
 static GObjectClass *parent_class = NULL;
@@ -130,6 +136,11 @@ gs_watcher_reset (GSWatcher *watcher)
 {
         g_return_if_fail (GS_IS_WATCHER (watcher));
 
+        /* just return quietly if not enabled */
+        if (! watcher->priv->enabled) {
+                return;
+        }
+
         /* restart if necessary */
         if (watcher->priv->active) {
                 gs_watcher_set_active (watcher, FALSE);
@@ -143,9 +154,12 @@ gs_watcher_set_timeout (GSWatcher  *watcher,
 {
         g_return_if_fail (GS_IS_WATCHER (watcher));
 
-        watcher->priv->timeout = timeout;
+        if (watcher->priv->timeout != timeout) {
+                watcher->priv->timeout = timeout;
 
-        gs_watcher_reset (watcher);
+                /* restart the timers if necessary */
+                gs_watcher_reset (watcher);
+        }
 }
 
 static void
@@ -199,16 +213,26 @@ gs_watcher_class_init (GSWatcherClass *klass)
         object_class->get_property = gs_watcher_get_property;
         object_class->set_property = gs_watcher_set_property;
 
-        signals [IDLE] =
-                g_signal_new ("idle",
+        signals [IDLE_CHANGED] =
+                g_signal_new ("idle_changed",
                               G_TYPE_FROM_CLASS (object_class),
                               G_SIGNAL_RUN_LAST,
-                              G_STRUCT_OFFSET (GSWatcherClass, idle),
+                              G_STRUCT_OFFSET (GSWatcherClass, idle_changed),
                               NULL,
                               NULL,
-                              gs_marshal_BOOLEAN__INT,
+                              gs_marshal_BOOLEAN__BOOLEAN,
                               G_TYPE_BOOLEAN,
-                              1, G_TYPE_INT);
+                              1, G_TYPE_BOOLEAN);
+        signals [IDLE_NOTICE_CHANGED] =
+                g_signal_new ("idle_notice_changed",
+                              G_TYPE_FROM_CLASS (object_class),
+                              G_SIGNAL_RUN_LAST,
+                              G_STRUCT_OFFSET (GSWatcherClass, idle_notice_changed),
+                              NULL,
+                              NULL,
+                              gs_marshal_BOOLEAN__BOOLEAN,
+                              G_TYPE_BOOLEAN,
+                              1, G_TYPE_BOOLEAN);
 
         g_object_class_install_property (object_class,
                                          PROP_TIMEOUT,
@@ -223,36 +247,10 @@ gs_watcher_class_init (GSWatcherClass *klass)
         g_type_class_add_private (klass, sizeof (GSWatcherPrivate));
 }
 
-static int
-saver_ehandler (Display     *dpy,
-                XErrorEvent *error)
-{
-        g_warning ("BUG BUG BUG!");
-        exit (1);
-}
-
-static int
-BadWindow_ehandler (Display     *dpy,
-                    XErrorEvent *error)
-{
-    /* When we notice a window being created, we spawn a timer that waits
-       30 seconds or so, and then selects events on that window.  This error
-       handler is used so that we can cope with the fact that the window
-       may have been destroyed <30 seconds after it was created.
-    */
-    if (error->error_code == BadWindow ||
-        error->error_code == BadMatch ||
-        error->error_code == BadDrawable)
-        return 0;
-    else
-        return saver_ehandler (dpy, error);
-}
-
 static void
 notice_events_inner (Window   window,
                      gboolean enable,
-                     gboolean top,
-                     gboolean debug)
+                     gboolean top)
 {
         XWindowAttributes attrs;
         unsigned long     events;
@@ -273,8 +271,9 @@ notice_events_inner (Window   window,
         status = XQueryTree (GDK_DISPLAY (), window, &root, &parent, &kids, &nkids);
 
         if (status == 0) {
-                if (kids)
+                if (kids) {
                         XFree (kids);
+                }
                 return;
         }
 
@@ -284,11 +283,28 @@ notice_events_inner (Window   window,
         XGetWindowAttributes (GDK_DISPLAY (), window, &attrs);
 
         if (enable) {
+                /* Select for KeyPress on all windows that already have it selected */
                 events = ((attrs.all_event_masks | attrs.do_not_propagate_mask) & KeyPressMask);
+
+                /* Keep already selected events.  This is important when the
+                   window == GDK_ROOT_WINDOW () since the mask will contain
+                   StructureNotifyMask that is essential for RANDR support */
+                events = attrs.your_event_mask | events;
+
+                /* Select for SubstructureNotify on all windows */
                 events = SubstructureNotifyMask | events;
+
+                /* Select for PropertyNotify events to get user time changes */
+                events = PropertyChangeMask | events;
         } else {
                 /* We want to disable all events */
-                events = 0;
+
+                /* Don't mess up the root window */
+                if (window == GDK_ROOT_WINDOW ()) {
+                        events = attrs.your_event_mask;
+                } else {
+                        events = 0;
+                }
         }
 
         /* Select for SubstructureNotify on all windows.
@@ -313,18 +329,20 @@ notice_events_inner (Window   window,
 
         XSelectInput (GDK_DISPLAY (), window, events);
 
-        if (top && debug && (events & KeyPressMask)) {
+        if (top && (events & KeyPressMask)) {
                 /* Only mention one window per tree */
                 top = FALSE;
-                if (enable)
-                        g_message ("Adding events for 0x%lX", (unsigned long)window);
-                else
-                        g_message ("Removing events for 0x%lX", (unsigned long)window);
+                if (enable) {
+                        gs_debug ("Adding events for 0x%lX", (unsigned long)window);
+                } else {
+                        gs_debug ("Removing events for 0x%lX", (unsigned long)window);
+                }
         }
 
         if (kids) {
-                while (nkids)
-                        notice_events_inner (kids [--nkids], enable, top, debug);
+                while (nkids) {
+                        notice_events_inner (kids [--nkids], enable, top);
+                }
 
                 XFree (kids);
         }
@@ -333,16 +351,14 @@ notice_events_inner (Window   window,
 static void
 notice_events (Window   window,
                gboolean enable,
-               gboolean top,
-               gboolean debug)
+               gboolean top)
 {
-        XErrorHandler old_handler;
+        gdk_error_trap_push ();
 
-        old_handler = XSetErrorHandler (BadWindow_ehandler);
+        notice_events_inner (window, enable, top);
 
-        notice_events_inner (window, enable, top, debug);
-        XSync (GDK_DISPLAY (), FALSE);
-        XSetErrorHandler (old_handler);
+        gdk_flush ();
+        gdk_error_trap_pop ();
 }
 
 typedef struct _NoticeTimerData NoticeTimerData;
@@ -357,11 +373,10 @@ static gboolean
 notice_events_timer (NoticeTimerData *data)
 {
         gboolean is_top = TRUE;
-        gboolean debug  = data->watcher->priv->debug;
         gboolean enable = data->enable;
         Window   window = data->window;
 
-        notice_events (window, enable, is_top, debug);
+        notice_events (window, enable, is_top);
 
         return FALSE;
 }
@@ -417,32 +432,129 @@ start_notice_events_timer (GSWatcher *watcher,
 
 }
 
+static void
+remove_idle_timer (GSWatcher *watcher)
+{
+        if (watcher->priv->timer_id != 0) {
+                gs_debug ("killing idle_timer  (%u, %u)",
+                          watcher->priv->timeout,
+                          watcher->priv->timer_id);
+
+                g_source_remove (watcher->priv->timer_id);
+                watcher->priv->timer_id = 0;
+        }
+}
+
+static void
+add_idle_timer (GSWatcher *watcher,
+                glong      timeout)
+{
+        watcher->priv->timer_id = g_timeout_add (timeout, (GSourceFunc)idle_timer, watcher);
+
+        gs_debug ("starting idle_timer (%ld, %u)", timeout, watcher->priv->timer_id);
+}
+
+static void
+remove_watchdog_timer (GSWatcher *watcher)
+{
+        if (watcher->priv->watchdog_timer_id != 0) {
+                g_source_remove (watcher->priv->watchdog_timer_id);
+                watcher->priv->watchdog_timer_id = 0;
+        }
+}
+
+static void
+add_watchdog_timer (GSWatcher *watcher,
+                    glong      timeout)
+{
+        watcher->priv->watchdog_timer_id = g_timeout_add (timeout,
+                                                          (GSourceFunc)watchdog_timer,
+                                                          watcher);
+}
+
 /* Call this when user activity (or "simulated" activity) has been noticed.
  */
 static void
 reset_timers (GSWatcher *watcher)
 {
 
-        if (watcher->priv->using_mit_saver_extension || watcher->priv->using_sgi_saver_extension)
+        if (watcher->priv->using_mit_saver_extension || watcher->priv->using_sgi_saver_extension) {
                 return;
-
-        if (watcher->priv->timer_id) {
-                if (watcher->priv->debug)
-                        g_message ("killing idle_timer  (%u, %u)", watcher->priv->timeout, watcher->priv->timer_id);
-                g_source_remove (watcher->priv->timer_id);
-                watcher->priv->timer_id = 0;
         }
 
-        schedule_wakeup_event (watcher, watcher->priv->timeout, watcher->priv->debug);
+        remove_idle_timer (watcher);
 
-        watcher->priv->last_activity_time = time (NULL);
+        schedule_wakeup_event (watcher, watcher->priv->timeout);
+
+        g_timer_start (watcher->priv->idle_timer);
+}
+
+static gboolean
+_gs_watcher_set_session_idle_notice (GSWatcher *watcher,
+                                     gboolean   in_effect)
+{
+        gboolean res;
+
+        res = FALSE;
+
+        if (in_effect != watcher->priv->idle_notice) {
+
+                g_signal_emit (watcher, signals [IDLE_NOTICE_CHANGED], 0, in_effect, &res);
+                if (res) {
+                        gs_debug ("Changing idle notice state: %d", in_effect);
+
+                        watcher->priv->idle_notice = in_effect;
+                } else {
+                        gs_debug ("Idle notice signal not handled: %d", in_effect);
+                }
+        }
+
+        return res;
+}
+
+static gboolean
+_gs_watcher_set_session_idle (GSWatcher *watcher,
+                              gboolean   is_idle)
+{
+        gboolean res;
+
+        res = FALSE;
+
+        if (is_idle != watcher->priv->idle) {
+
+                g_signal_emit (watcher, signals [IDLE_CHANGED], 0, is_idle, &res);
+                if (res) {
+                        gs_debug ("Changing idle state: %d", is_idle);
+
+                        watcher->priv->idle = is_idle;
+                } else {
+                        gs_debug ("Idle changed signal not handled: %d", is_idle);
+                }
+        }
+
+        return res;
 }
 
 static void
 _gs_watcher_notice_activity (GSWatcher *watcher)
 {
-        if (watcher->priv->debug) {
-                g_message ("Activity detected: resetting timers");
+        if (! watcher->priv->active) {
+                gs_debug ("Noticed activity but watcher is inactive");
+                return;
+        }
+
+        gs_debug ("Activity detected: resetting timers");
+
+        /* if an idle notice was sent, cancel it */
+        if (watcher->priv->idle_notice) {
+                gboolean in_effect = FALSE;
+                _gs_watcher_set_session_idle_notice (watcher, in_effect);
+        }
+
+        /* if idle signal was sent, cancel it */
+        if (watcher->priv->idle) {
+                gboolean is_idle = FALSE;
+                _gs_watcher_set_session_idle (watcher, is_idle);
         }
 
         reset_timers (watcher);
@@ -452,9 +564,7 @@ static void
 _gs_watcher_notice_window_created (GSWatcher *watcher,
                                    Window     window)
 {
-        if (watcher->priv->debug) {
-                g_message ("Window created: noticing activity on 0x%lX", (unsigned long)window);
-        }
+        gs_debug ("Window created: noticing activity on 0x%lX", (unsigned long)window);
 
         start_notice_events_timer (watcher, window);
 }
@@ -466,8 +576,9 @@ gs_watcher_xevent (GSWatcher *watcher,
         XEvent *ev;
 
         /* do nothing if we aren't watching */
-        if (! watcher->priv->active)
+        if (! watcher->priv->active) {
                 return;
+        }
 
         ev = xevent;
 
@@ -478,6 +589,11 @@ gs_watcher_xevent (GSWatcher *watcher,
         case ButtonRelease:
                 _gs_watcher_notice_activity (watcher);
                 break;
+        case PropertyNotify:
+                if (ev->xproperty.atom == gdk_x11_get_xatom_by_name ("_NET_WM_USER_TIME")) {
+                        _gs_watcher_notice_activity (watcher);
+                }
+                break;
         case CreateNotify:
                 {
                         Window window = ev->xcreatewindow.window;
@@ -501,18 +617,33 @@ xevent_filter (GdkXEvent *xevent,
         return GDK_FILTER_CONTINUE;
 }
 
+static void
+remove_check_pointer_timer (GSWatcher *watcher)
+{
+        if (watcher->priv->check_pointer_timer_id != 0) {
+                g_source_remove (watcher->priv->check_pointer_timer_id);
+                watcher->priv->check_pointer_timer_id = 0;
+        }
+}
+
+static void
+add_check_pointer_timer (GSWatcher *watcher,
+                         glong      timeout)
+{
+        watcher->priv->check_pointer_timer_id = g_timeout_add (timeout,
+                                                               (GSourceFunc)check_pointer_timer, watcher);
+
+}
+
 static void
 start_pointer_poll (GSWatcher *watcher)
 {
         /* run once to set baseline */
         check_pointer_timer (watcher);
 
-        if (watcher->priv->check_pointer_timer_id) {
-                g_source_remove (watcher->priv->check_pointer_timer_id);
-                watcher->priv->check_pointer_timer_id = 0;
-        }
-        watcher->priv->check_pointer_timer_id = g_timeout_add (watcher->priv->pointer_timeout,
-                                                               (GSourceFunc)check_pointer_timer, watcher);
+        remove_check_pointer_timer (watcher);
+
+        add_check_pointer_timer (watcher, watcher->priv->pointer_timeout);
 }
 
 static void
@@ -538,10 +669,7 @@ _gs_watcher_set_pointer_position (GSWatcher       *watcher,
 static void
 stop_pointer_poll (GSWatcher *watcher)
 {
-        if (watcher->priv->check_pointer_timer_id != 0) {
-                g_source_remove (watcher->priv->check_pointer_timer_id);
-                watcher->priv->check_pointer_timer_id = 0;
-        }
+        remove_check_pointer_timer (watcher);
 
         _gs_watcher_set_pointer_position (watcher, NULL);
 }
@@ -552,8 +680,8 @@ start_idle_watcher (GSWatcher *watcher)
         g_return_val_if_fail (watcher != NULL, FALSE);
         g_return_val_if_fail (GS_IS_WATCHER (watcher), FALSE);
 
-        watcher->priv->last_wall_clock_time = 0;
-        watcher->priv->last_activity_time = time (NULL);
+        g_timer_start (watcher->priv->jump_timer);
+        g_timer_start (watcher->priv->idle_timer);
 
         start_pointer_poll (watcher);
 
@@ -571,13 +699,10 @@ stop_idle_watcher (GSWatcher *watcher)
         g_return_val_if_fail (watcher != NULL, FALSE);
         g_return_val_if_fail (GS_IS_WATCHER (watcher), FALSE);
 
-        watcher->priv->last_wall_clock_time = 0;
-        watcher->priv->last_activity_time = time (NULL);
+        g_timer_stop (watcher->priv->jump_timer);
+        g_timer_stop (watcher->priv->idle_timer);
 
-        if (watcher->priv->timer_id != 0) {
-                g_source_remove (watcher->priv->timer_id);
-                watcher->priv->timer_id = 0;
-        }
+        remove_idle_timer (watcher);
 
         stop_pointer_poll (watcher);
 
@@ -599,48 +724,98 @@ gs_watcher_get_active (GSWatcher *watcher)
         return active;
 }
 
+static gboolean
+_gs_watcher_set_active_internal (GSWatcher *watcher,
+                                 gboolean   active)
+{
+        if (! active) {
+                watcher->priv->active = FALSE;
+                stop_idle_watcher (watcher);
+                gs_debug ("Stopping idle watcher");
+        } else {
+                watcher->priv->active = TRUE;
+                start_idle_watcher (watcher);
+                gs_debug ("Starting idle watcher");
+        }
+
+        return TRUE;
+}
+
+
 gboolean
 gs_watcher_set_active (GSWatcher *watcher,
                        gboolean   active)
 {
         g_return_val_if_fail (GS_IS_WATCHER (watcher), FALSE);
 
+        gs_debug ("turning watcher: %s", active ? "ON" : "OFF");
+
         if (watcher->priv->active == active) {
-                g_warning ("Watcher is already %s",
-                           active ? "active" : "inactive");
+                gs_debug ("Idle detection is already %s",
+                          active ? "active" : "inactive");
                 return FALSE;
         }
 
-        if (! active) {
-                watcher->priv->active = FALSE;
-                stop_idle_watcher (watcher);
-                if (watcher->priv->debug)
-                        g_message ("Stopping idle watcher");
-        } else {
-                watcher->priv->active = TRUE;
-                start_idle_watcher (watcher);
-                if (watcher->priv->debug)
-                        g_message ("Starting idle watcher");
+        if (! watcher->priv->enabled) {
+                gs_debug ("Idle detection is disabled, cannot activate");
+                return FALSE;
+        }
+
+        return _gs_watcher_set_active_internal (watcher, active);
+}
+
+gboolean
+gs_watcher_set_enabled (GSWatcher *watcher,
+                        gboolean   enabled)
+{
+        g_return_val_if_fail (GS_IS_WATCHER (watcher), FALSE);
+
+        if (watcher->priv->enabled != enabled) {
+                gboolean is_active = gs_watcher_get_active (watcher);
+
+                watcher->priv->enabled = enabled;
+
+                /* if we are disabling the watcher and we are
+                   active shut it down */
+                if (! enabled && is_active) {
+                        _gs_watcher_set_active_internal (watcher, FALSE);
+                }
         }
 
         return TRUE;
 }
 
+gboolean
+gs_watcher_get_enabled (GSWatcher *watcher)
+{
+        gboolean enabled;
+
+        g_return_val_if_fail (GS_IS_WATCHER (watcher), FALSE);
+
+        enabled = watcher->priv->enabled;
+
+        return enabled;
+}
+
 static void
 gs_watcher_init (GSWatcher *watcher)
 {
         watcher->priv = GS_WATCHER_GET_PRIVATE (watcher);
 
+        watcher->priv->enabled = TRUE;
         watcher->priv->active = FALSE;
         watcher->priv->timeout = 600000;
-        watcher->priv->pointer_timeout = 5000;
+        watcher->priv->pointer_timeout = 1000;
 
-        watcher->priv->verbose = FALSE;
-        watcher->priv->debug = FALSE;
+        /* time before idle signal to send notice signal */
+        watcher->priv->notice_timeout = 10000;
+
+        watcher->priv->idle_timer = g_timer_new ();
+        watcher->priv->jump_timer = g_timer_new ();
 
         initialize_server_extensions (watcher);
 
-        watcher->priv->watchdog_timer_id = g_timeout_add (600000, (GSourceFunc)watchdog_timer, watcher);
+        add_watchdog_timer (watcher, 600000);
 }
 
 static void
@@ -655,16 +830,18 @@ gs_watcher_finalize (GObject *object)
 
         g_return_if_fail (watcher->priv != NULL);
 
-        if (watcher->priv->watchdog_timer_id != 0) {
-                g_source_remove (watcher->priv->watchdog_timer_id);
-                watcher->priv->watchdog_timer_id = 0;
-        }
+        remove_watchdog_timer (watcher);
 
         watcher->priv->active = FALSE;
         stop_idle_watcher (watcher);
 
         _gs_watcher_pointer_position_free (watcher->priv->poll_position);
 
+        g_timer_destroy (watcher->priv->idle_timer);
+        watcher->priv->idle_timer = NULL;
+        g_timer_destroy (watcher->priv->jump_timer);
+        watcher->priv->jump_timer = NULL;
+
         G_OBJECT_CLASS (parent_class)->finalize (object);
 }
 
@@ -681,14 +858,6 @@ query_mit_saver_extension (int *event_number,
                                            error_number);
 }
 
-static int
-ignore_all_errors_ehandler (Display     *dpy,
-                            XErrorEvent *error)
-{
-        return 0;
-}
-
-
 /* MIT SCREEN-SAVER server extension hackery.
  */
 static gboolean
@@ -714,12 +883,11 @@ init_mit_saver_extension ()
                                                XScreenNumberOfScreen (GDK_SCREEN_XSCREEN (screen)),
                                                &kill_id, &kill_type)
                     && kill_id != blank_pix[i]) {
-                        XErrorHandler old_handler =
-                                XSetErrorHandler (ignore_all_errors_ehandler);
+                        gdk_error_trap_push ();
 
                         XKillClient (GDK_DISPLAY (), kill_id);
-                        XSync (GDK_DISPLAY (), FALSE);
-                        XSetErrorHandler (old_handler);
+                        gdk_flush ();
+                        gdk_error_trap_pop ();
                 }
 
                 XScreenSaverSelectInput (GDK_DISPLAY (), root, ScreenSaverNotifyMask);
@@ -769,7 +937,8 @@ init_sgi_saver_extension (gboolean is_blank_active)
 
                 XScreenSaverDisable (GDK_DISPLAY (), XScreenNumberOfScreen (GDK_SCREEN_XSCREEN (screen)));
                 if (! XScreenSaverEnable (GDK_DISPLAY (), XScreenNumberOfScreen (GDK_SCREEN_XSCREEN (screen)))) {
-                        g_message ("%SGI SCREEN_SAVER extension exists, but can't be initialized;\n"
+
+                        g_warning ("%SGI SCREEN_SAVER extension exists, but can't be initialized;\n"
                                    "perhaps some other screensaver program is already running?");
 
                         return FALSE;
@@ -826,38 +995,41 @@ initialize_server_extensions (GSWatcher *watcher)
                                                                     &watcher->priv->mit_saver_ext_error_number);
 #endif
 
-        if (! server_has_xidle_extension)
+        if (! server_has_xidle_extension) {
                 watcher->priv->using_xidle_extension = FALSE;
-        else if (watcher->priv->verbose) {
-                if (watcher->priv->using_xidle_extension)
-                        g_message ("Using XIDLE extension.");
-                else
-                        g_message ("Not using server's XIDLE extension.");
+        } else {
+                if (watcher->priv->using_xidle_extension) {
+                        gs_debug ("Using XIDLE extension.");
+                } else {
+                        gs_debug ("Not using server's XIDLE extension.");
                 }
-
-        if (! server_has_sgi_saver_extension)
-                watcher->priv->using_sgi_saver_extension = FALSE;
-        else if (watcher->priv->verbose) {
-                if (watcher->priv->using_sgi_saver_extension)
-                        g_message ("Using SGI SCREEN_SAVER extension.");
-                else
-                        g_message ("Not using server's SGI SCREEN_SAVER extension.");
         }
 
-        if (! server_has_mit_saver_extension)
+        if (! server_has_sgi_saver_extension) {
+                watcher->priv->using_sgi_saver_extension = FALSE;
+        } else {
+                if (watcher->priv->using_sgi_saver_extension) {
+                        gs_debug ("Using SGI SCREEN_SAVER extension.");
+                } else {
+                        gs_debug ("Not using server's SGI SCREEN_SAVER extension.");
+                }
+        }
+
+        if (! server_has_mit_saver_extension) {
                 watcher->priv->using_mit_saver_extension = FALSE;
-        else if (watcher->priv->verbose) {
-                if (watcher->priv->using_mit_saver_extension)
-                        g_message ("Using MIT-SCREEN-SAVER extension.");
-                else
-                        g_message ("Not using server's MIT-SCREEN-SAVER extension.");
+        } else {
+                if (watcher->priv->using_mit_saver_extension) {
+                        gs_debug ("Using MIT-SCREEN-SAVER extension.");
+                } else {
+                        gs_debug ("Not using server's MIT-SCREEN-SAVER extension.");
+                }
         }
 
         /* These are incompatible (or at least, our support for them is...) */
         if (watcher->priv->xinerama && watcher->priv->using_mit_saver_extension) {
                 watcher->priv->using_mit_saver_extension = FALSE;
-                if (watcher->priv->verbose)
-                        g_message ("Xinerama in use: disabling MIT-SCREEN-SAVER.");
+
+                gs_debug ("Xinerama in use: disabling MIT-SCREEN-SAVER.");
         }
 }
 
@@ -874,8 +1046,11 @@ disable_builtin_screensaver (GSWatcher *watcher,
         int desired_server_timeout, desired_server_interval;
         int desired_prefer_blank,   desired_allow_exp;
 
-        XGetScreenSaver (GDK_DISPLAY (), &current_server_timeout, &current_server_interval,
-                         &current_prefer_blank, &current_allow_exp);
+        XGetScreenSaver (GDK_DISPLAY (),
+                         &current_server_timeout,
+                         &current_server_interval,
+                         &current_prefer_blank,
+                         &current_allow_exp);
 
         desired_server_timeout  = current_server_timeout;
         desired_server_interval = current_server_interval;
@@ -925,17 +1100,19 @@ disable_builtin_screensaver (GSWatcher *watcher,
             || desired_server_interval != current_server_interval
             || desired_prefer_blank    != current_prefer_blank
             || desired_allow_exp       != current_allow_exp) {
- 
-               if (watcher->priv->verbose)
-                        g_message ("disabling server builtin screensaver:"
-                                   " (xset s %d %d; xset s %s; xset s %s)",
-                                   desired_server_timeout, desired_server_interval,
-                                   (desired_prefer_blank ? "blank" : "noblank"),
-                                   (desired_allow_exp ? "expose" : "noexpose"));
+
+                gs_debug ("disabling server builtin screensaver:"
+                          " (xset s %d %d; xset s %s; xset s %s)",
+                          desired_server_timeout,
+                          desired_server_interval,
+                          (desired_prefer_blank ? "blank" : "noblank"),
+                          (desired_allow_exp ? "expose" : "noexpose"));
 
                 XSetScreenSaver (GDK_DISPLAY (),
-                                 desired_server_timeout, desired_server_interval,
-                                 desired_prefer_blank, desired_allow_exp);
+                                 desired_server_timeout,
+                                 desired_server_interval,
+                                 desired_prefer_blank,
+                                 desired_allow_exp);
 
                 XSync (GDK_DISPLAY (), FALSE);
         }
@@ -974,18 +1151,30 @@ static void
 maybe_send_signal (GSWatcher *watcher)
 {
         gboolean polling_for_idleness = TRUE;
-        int      idle;
-        gboolean do_signal = FALSE;
+        gint64   elapsed;
+        gboolean do_idle_signal = FALSE;
+        gboolean do_notice_signal = FALSE;
 
-        idle = 1000 * (time (NULL) - watcher->priv->last_activity_time);
+        if (! watcher->priv->active) {
+                gs_debug ("Checking for idleness but watcher is inactive");
+                return;
+        }
 
-        if (idle >= watcher->priv->timeout) {
+        if (watcher->priv->idle) {
+                /* already idle, do nothing */
+                return;
+        }
+
+        elapsed = 1000 * g_timer_elapsed (watcher->priv->idle_timer, NULL);
+
+        if (elapsed >= watcher->priv->timeout) {
                 /* Look, we've been idle long enough.  We're done. */
-                do_signal = TRUE;
+                do_idle_signal = TRUE;
         } else if (watcher->priv->emergency_lock) {
                 /* Oops, the wall clock has jumped far into the future, so
                    we need to lock down in a hurry! */
-                do_signal = TRUE;
+                gs_debug ("Doing emergency lock");
+                do_idle_signal = TRUE;
         } else {
                 /* The event went off, but it turns out that the user has not
                    yet been idle for long enough.  So re-signal the event.
@@ -993,22 +1182,40 @@ maybe_send_signal (GSWatcher *watcher)
                    user has been idle for 2 minutes, then set this timer to
                    go off in 3 minutes.
                 */
-                if (polling_for_idleness)
-                        schedule_wakeup_event (watcher, watcher->priv->timeout - idle, watcher->priv->debug);
-                do_signal = FALSE;
+
+                if (polling_for_idleness) {
+                        guint time_left;
+
+                        time_left = watcher->priv->timeout - elapsed;
+
+                        if (time_left <= watcher->priv->notice_timeout) {
+                                do_notice_signal = TRUE;
+                        }
+
+                        schedule_wakeup_event (watcher, time_left);
+                }
+
+                do_idle_signal = FALSE;
         }
 
-        if (do_signal) {
+        if (do_notice_signal && ! watcher->priv->idle_notice) {
 		gboolean res = FALSE;
+                gboolean in_effect = TRUE;
 
-                g_signal_emit (watcher, signals [IDLE], 0, 0, &res);
+                res = _gs_watcher_set_session_idle_notice (watcher, in_effect);
+        }
+
+        if (do_idle_signal) {
+		gboolean res = FALSE;
+                gboolean is_idle = TRUE;
+
+                res = _gs_watcher_set_session_idle (watcher, is_idle);
+                _gs_watcher_set_session_idle_notice (watcher, !is_idle);
 
                 /* if the event wasn't handled then schedule another timer */
                 if (! res) {
-                        g_warning ("Idle signal was not handled, restarting watcher");
-
-                        if (polling_for_idleness)
-                                schedule_wakeup_event (watcher, watcher->priv->timeout, watcher->priv->debug);
+                        gs_debug ("Idle signal was not handled, restarting watcher");
+                        gs_watcher_reset (watcher);
                 }
         }
 }
@@ -1016,6 +1223,8 @@ maybe_send_signal (GSWatcher *watcher)
 static gboolean
 idle_timer (GSWatcher *watcher)
 {
+        gs_debug ("in idle timer");
+
         /* try one last time */
         check_pointer_timer (watcher);
 
@@ -1028,21 +1237,24 @@ idle_timer (GSWatcher *watcher)
 
 static void
 schedule_wakeup_event (GSWatcher *watcher,
-                       int        when,
-                       gboolean   verbose)
+                       int        when)
 {
+        guint timeout;
+
         if (watcher->priv->timer_id) {
-                if (verbose)
-                        g_message ("idle_timer already running");
+                gs_debug ("idle_timer already running");
                 return;
         }
 
+        timeout = when;
+
+        /* Wake up before idle so we can send a notice signal */
+        if (timeout > watcher->priv->notice_timeout) {
+                timeout -= watcher->priv->notice_timeout;
+        }
+
         /* Wake up periodically to ask the server if we are idle. */
-        watcher->priv->timer_id = g_timeout_add (when, (GSourceFunc)idle_timer, watcher);
-
-
-        if (verbose)
-                g_message ("starting idle_timer (%d, %d)", when, watcher->priv->timer_id);
+        add_idle_timer (watcher, timeout);
 }
 
 /* An unfortunate situation is this: the saver is not active, because the
@@ -1065,25 +1277,41 @@ schedule_wakeup_event (GSWatcher *watcher,
 static void
 check_for_clock_skew (GSWatcher *watcher)
 {
-        time_t now = time ((time_t *) 0);
-        long shift = now - watcher->priv->last_wall_clock_time;
+        time_t now;
+        gint64 shift;
+        gint64 i;
+        gint64 i_h;
+        gint64 i_m;
+        gint64 i_s;
 
-        if (watcher->priv->debug) {
-                int i = (watcher->priv->last_wall_clock_time == 0 ? 0 : shift);
-                g_message ("checking wall clock for hibernation (%d:%02d:%02d).",
-                           (i / (60 * 60)), ((i / 60) % 60), (i % 60));
-        }
+        now   = time (NULL);
+        shift = g_timer_elapsed (watcher->priv->jump_timer, NULL);
+        i     = shift;
 
-        if (watcher->priv->last_wall_clock_time != 0 && shift > (watcher->priv->timeout / 1000)) {
-                if (watcher->priv->verbose)
-                        g_message ("wall clock has jumped by %ld:%02ld:%02ld!",
-                                   (shift / (60 * 60)), ((shift / 60) % 60), (shift % 60));
+        i_h = i / (60 * 60);
+        i_m = (i / 60) % 60;
+        i_s = i % 60;
+        gs_debug ("checking wall clock for hibernation, changed: %" G_GINT64_FORMAT ":%02" G_GINT64_FORMAT ":%02" G_GINT64_FORMAT,
+                  i_h, i_m, i_s);
+
+        if (shift > (watcher->priv->timeout / 1000)) {
+                gint64 s_h;
+                gint64 s_m;
+                gint64 s_s;
+
+                s_h = shift / (60 * 60);
+                s_m = (shift / 60) % 60;
+                s_s = shift % 60;
+
+                gs_debug ("wall clock has jumped by %" G_GINT64_FORMAT ":%02" G_GINT64_FORMAT ":%02" G_GINT64_FORMAT,
+                          s_h, s_m, s_s);
 
                 watcher->priv->emergency_lock = TRUE;
                 maybe_send_signal (watcher);
+                watcher->priv->emergency_lock = FALSE;
         }
 
-        watcher->priv->last_wall_clock_time = now;
+        g_timer_start (watcher->priv->jump_timer);
 }
 
 static PointerPosition *
@@ -1129,6 +1357,7 @@ _gs_watcher_check_pointer_position (GSWatcher *watcher)
 {
         PointerPosition *pos;
         gboolean         changed;
+        gint64           elapsed;
 
         pos = _gs_watcher_pointer_position_read (watcher);
 
@@ -1138,10 +1367,12 @@ _gs_watcher_check_pointer_position (GSWatcher *watcher)
         if (changed) {
                 _gs_watcher_set_pointer_position (watcher, pos);
                 _gs_watcher_notice_activity (watcher);
+        } else {
+                _gs_watcher_pointer_position_free (pos);
         }
 
-        if (watcher->priv->debug)
-                g_message ("Idle %d seconds", (int)(time (NULL) - watcher->priv->last_activity_time));
+        elapsed = g_timer_elapsed (watcher->priv->idle_timer, NULL);
+        gs_debug ("Idle %" G_GINT64_FORMAT " seconds", elapsed);
 
         check_for_clock_skew (watcher);
 }
