@@ -152,9 +152,9 @@ gs_window_override_user_time (GSWindow *window)
                                                GDK_PROPERTY_CHANGE_MASK);
                 }
 
-                /* 
+                /*
                  * NOTE: Last resort for D-BUS or other non-interactive
-                 *       openings.  Causes roundtrip to server.  Lame. 
+                 *       openings.  Causes roundtrip to server.  Lame.
                  */
                 ev_time = gdk_x11_get_server_time (GTK_WIDGET (window)->window);
         }
@@ -576,7 +576,8 @@ get_env_vars (GtkWidget *widget)
                 "KRB5CCNAME",
                 "KRBTKFILE",
                 "LANG",
-                "LANGUAGE"
+                "LANGUAGE",
+                "RUNNING_UNDER_GDM"
         };
 
         env = g_ptr_array_new ();
@@ -808,7 +809,7 @@ gs_window_dialog_finish (GSWindow *window)
 
         if (window->priv->pid > 0) {
                 int exit_status;
-                        
+
                 exit_status = wait_on_child (window->priv->pid);
 
                 g_spawn_close_pid (window->priv->pid);
@@ -983,6 +984,10 @@ popup_dialog_idle (GSWindow *window)
                 command = g_string_append (command, " --enable-switch");
         }
 
+        if (gs_debug_enabled ()) {
+                command = g_string_append (command, " --verbose");
+        }
+
         gs_window_clear (window);
         set_invisible_cursor (GTK_WIDGET (window)->window, FALSE);
 
@@ -1280,7 +1285,7 @@ gs_window_real_motion_notify_event (GtkWidget      *widget,
 {
         GSWindow *window;
         gdouble   distance;
-        gdouble   min_distance = 10;
+        gdouble   min_distance = 50;
 
         window = GS_WINDOW (widget);
 
